#include <stdatomic.h>
#include <unrelacy.h>
#include <common.h>

#include <spec_lib.h>
#include <stdlib.h>
#include <cdsannotate.h>
#include <specannotation.h>
#include <model_memory.h>

/**
	@Begin
	@Class_begin
	@End
*/
template <typename t_element, size_t t_size>
struct mpmc_boundq_1_alt
{
private:
	
	unsigned int MASK;

	// elements should generally be cache-line-size padded :
	t_element               m_array[t_size];

	// rdwr counts the reads & writes that have started
	atomic<unsigned int>    m_rdwr;
	// "read" and "written" count the number completed
	atomic<unsigned int>    m_read;
	atomic<unsigned int>    m_written;

public:

	mpmc_boundq_1_alt()
	{
    	/**
	    	@Begin
			@Entry_point
			@End
		*/
		m_rdwr = 0;
		m_read = 0;
		m_written = 0;
		// For this we want MASK = 1; MASK wrap around
 		MASK = 0x1; // 11
	}
	

	/**
		@Begin
		@Options:
			LANG = CPP;
			CLASS = mpmc_boundq_1_alt;
		@Global_define:
		@Happens_before:
			Publish -> Fetch
			Consume -> Prepare
		@Commutativity: Prepare <-> Prepare: _Method1.__RET__ !=
		_Method2.__RET__ || !_Method1.__RET__ || !_Method2.__RET__
		@Commutativity: Prepare <-> Publish: _Method1.__RET__ != _Method2.bin ||
		!_Method1.__RET__
		@Commutativity: Prepare <-> Fetch: _Method1.__RET__ != _Method2.__RET__
		|| !_Method1.__RET__ || !_Method2.__RET__
		@Commutativity: Prepare <-> Consume : _Method1.__RET__ != _Method2.bin || !_Method1.__RET__

		@Commutativity: Publish <-> Publish: _Method1.bin != _Method2.bin
		@Commutativity: Publish <-> Fetch: _Method1.bin != _Method2.__RET__ ||
		!_Method2.__RET__
		@Commutativity: Publish <-> Consume : _Method1.bin != _Method2.bin

		@Commutativity: Fetch <-> Fetch: _Method1.__RET__ != _Method2.__RET__ ||
		!_Method1.__RET__ || !_Method2.__RET__
		@Commutativity: Fetch <-> Consume : _Method1.__RET__ != _Method2.bin || !_Method1.__RET__

		@Commutativity: Consume <-> Consume : _Method1.bin != _Method2.bin

	@End
	*/

	//-----------------------------------------------------

	/**
		@Begin
		@Interface: Fetch
		@Commit_point_set: Fetch_RW_Load_Empty | Fetch_RW_RMW | Fetch_W_Load
		@ID: (call_id_t) __RET__
		//@Action: model_print("Fetch: %d\n", __RET__);
		@End
	*/
	t_element * read_fetch() {
		// Since we have a lool to CAS the value of m_rdwr, this can be relaxed
		unsigned int rdwr = m_rdwr.load(mo_acquire);
		//unsigned int rdwr = m_rdwr.load(mo_relaxed);
		/**
			@Begin
			@Potential_commit_point_define: true
			@Label: Fetch_Potential_RW_Load
			@End
		*/
		unsigned int rd,wr;
		for(;;) {
			rd = (rdwr>>16) & MASK;
			wr = rdwr & MASK;

			if ( wr == rd ) { // empty

				/**
					@Begin
					@Commit_point_define: true
					@Potential_commit_point_label: Fetch_Potential_RW_Load
					@Label: Fetch_RW_Load_Empty
					@End
				*/

				return false;
			}
			
			/**** Inadmissibility (testcase2.cc, MASK = 1, size = 1) ****/
			bool succ =
			m_rdwr.compare_exchange_weak(rdwr,rdwr+(1<<16),mo_acq_rel);
			/**
				@Begin
				@Commit_point_define_check: succ
				@Label: Fetch_RW_RMW
				@End
			*/
			if (succ)
				break;
			else
				thrd_yield();
		}

		// (*1)
		rl::backoff bo;
		while (true) {
			/**** Inadmissibility ****/
			int written = m_written.load(mo_acquire);
			/**
				@Begin
				@Potential_commit_point_define: true
				@Label: Fetch_Potential_W_Load
				@End
			*/
			if ((written & MASK) != wr) {
				thrd_yield();
			} else {
				printf("Fetch: m_written=%d\n", written);
				break;
			}
		}
		
		/**
			@Begin
			@Commit_point_define: true
			@Potential_commit_point_label: Fetch_Potential_W_Load
			@Label: Fetch_W_Load
			@End
		*/

		t_element * p = & ( m_array[ rd % t_size ] );
		
		return p;
	}

	/**
		@Begin
		@Interface: Consume
		@Commit_point_set: Consume_R_RMW
		@ID: (call_id_t) bin 
		//@Action: model_print("Consume: %d\n", bin);
		@End
	*/
	void read_consume(t_element *bin) {
		/**** Inadmissibility ****/
		m_read.fetch_add(1,mo_release);
		/**
			@Begin
			@Commit_point_define_check: true
			@Label: Consume_R_RMW
			@End
		*/
	}

	//-----------------------------------------------------

	/**
		@Begin
		@Interface: Prepare 
		@Commit_point_set: Prepare_RW_Load_Full | Prepare_RW_RMW | Prepare_R_Load
		@ID: (call_id_t) __RET__
		//@Action: model_print("Prepare: %d\n", __RET__);
		@End
	*/
	t_element * write_prepare() {
		// Try weaker semantics
		// Since we have a lool to CAS the value of m_rdwr, this can be relaxed
		unsigned int rdwr = m_rdwr.load(mo_acquire);
		//unsigned int rdwr = m_rdwr.load(mo_relaxed);
		/**
			@Begin
			@Potential_commit_point_define: true
			@Label: Prepare_Potential_RW_Load
			@End
		*/
		unsigned int rd,wr;
		for(;;) {
			rd = (rdwr>>16) & MASK;
			wr = rdwr & MASK;
			//printf("write_prepare: rd=%d, wr=%d\n", rd, wr);

			if ( wr == ((rd + t_size)&MASK) ) { // full

				/**
					@Begin
					@Commit_point_define: true
					@Potential_commit_point_label: Prepare_Potential_RW_Load
					@Label: Prepare_RW_Load_Full
					@End
				*/
				return NULL;
			}
			
			/**** Inadmissibility (testcase3.cc, MASK = 1, size = 1) ****/
			bool succ = m_rdwr.compare_exchange_weak(rdwr,(rd<<16) |
				((wr+1)&MASK),mo_acq_rel);
			/**
				@Begin
				@Commit_point_define_check: succ
				@Label: Prepare_RW_RMW
				@End
			*/
			//printf("wr=%d\n", (wr+1)&MASK);
			if (succ)
				break;
			else
				thrd_yield();
		}

		// (*1)
		rl::backoff bo;
		while (true) {
			/**** Inadmissibility ****/
			int read = m_read.load(mo_acquire);
			/**
				@Begin
				@Potential_commit_point_define: true
				@Label: Prepare_Potential_R_Load
				@End
			*/
			if ((read & MASK) != rd)
				thrd_yield();
			else
				break;
		}

		/**
			@Begin
			@Commit_point_define: true
			@Potential_commit_point_label: Prepare_Potential_R_Load
			@Label: Prepare_R_Load
			@End
		*/

		t_element * p = & ( m_array[ wr % t_size ] );

		return p;
	}

	/**
		@Begin
		@Interface: Publish 
		@Commit_point_set: Publish_W_RMW
		@ID: (call_id_t) bin 
		//@Action: model_print("Publish: %d\n", bin);
		@End
	*/
	void write_publish(t_element *bin)
	{
		/**** Inadmissibility ****/
		int tmp = m_written.fetch_add(1,mo_release);
		/**
			@Begin
			@Commit_point_define_check: true
			@Label: Publish_W_RMW
			@End
		*/
		printf("publish: m_written=%d\n", tmp + 1);
	}

	//-----------------------------------------------------


};
/**
	@Begin
	@Class_end
	@End
*/
